require 'spec_helper'

unless defined?(VERTICA_DB)
  VERTICA_URL = 'vertica://vertica:vertica@localhost:5432/reality_spec' unless defined? VERTICA_URL
  VERTICA_DB = Sequel.connect(ENV['SEQUEL_VERTICA_SPEC_DB']||VERTICA_URL)
end
INTEGRATION_DB = VERTICA_DB unless defined?(INTEGRATION_DB)

def VERTICA_DB.sqls
  (@sqls ||= [])
end


VERTICA_DB.create_table! :test do
  varchar :name
  integer :value
end
VERTICA_DB.create_table! :test2 do
  varchar :name
  integer :value
end
VERTICA_DB.create_table! :test3 do
  integer :value
  timestamp :time
end
VERTICA_DB.create_table! :test4 do
  varchar :name, :size => 20
  bytea :value
end

describe "A vertica sequel connection" do
  specify "should set a read timeout" do
    conn = Sequel.connect("#{ENV['SEQUEL_VERTICA_SPEC_DB']||VERTICA_URL}?read_timeout=1000")
    conn.synchronize do |raw_conn|
      expect(raw_conn.options[:read_timeout]).to eq(1000)
    end
  end
end

describe "A Vertica database" do

  before do
    @db = VERTICA_DB
  end

  specify "should correctly parse the schema" do
    expect(@db.schema(:test3, :reload=>true)).to eq([
      [:value, {:type=>:integer, :allow_null=>true, :default=>nil, :ruby_default=>nil, :db_type=>"int", :primary_key=>false}],
      [:time, {:type=>:datetime, :allow_null=>true, :default=>nil, :ruby_default=>nil, :db_type=>"timestamp", :primary_key=>false}]
    ])
    expect(@db.schema(:test4, :reload=>true)).to eq([
      [:name, {:allow_null=>true, :default=>nil, :db_type=>"varchar(20)", :type=>:string, :primary_key=>false, :ruby_default=>nil, :max_length=>20}],
      [:value, {:allow_null=>true, :default=>nil, :db_type=>"varbinary(80)", :type=>:blob, :primary_key=>false, :ruby_default=>nil}]
    ])
  end

  specify "should create an auto incrementing primary key" do
    @db.create_table! :auto_inc_test do
      primary_key :id
      integer :value
    end
    expect(@db[<<-SQL].first[:COUNT]).to eq(1)
      SELECT COUNT(1) FROM v_catalog.sequences WHERE identity_table_name='auto_inc_test'
    SQL
  end

end

describe "A vertica dataset" do
  before do
    @d = VERTICA_DB[:test]
    @d.delete if @d.count > 0 # Vertica will throw an error if the table has just been created and does not have a super projection yet.
  end

  specify "should quote columns and tables using double quotes if quoting identifiers" do
    expect(@d.select(:name).sql).to eq( \
      'SELECT "name" FROM "test"'
    )

    expect(@d.select(Sequel.lit('COUNT(*)')).sql).to eq( \
      'SELECT COUNT(*) FROM "test"'
    )

    expect(@d.select(:max.sql_function(:value)).sql).to eq( \
      'SELECT max("value") FROM "test"'
    )

    expect(@d.select(:NOW.sql_function).sql).to eq( \
    'SELECT NOW() FROM "test"'
    )

    expect(@d.select(:max.sql_function(:items__value)).sql).to eq( \
      'SELECT max("items"."value") FROM "test"'
    )

    expect(@d.order(:name.desc).sql).to eq( \
      'SELECT * FROM "test" ORDER BY "name" DESC'
    )

    expect(@d.select(Sequel.lit('test.name AS item_name')).sql).to eq( \
      'SELECT test.name AS item_name FROM "test"'
    )

    expect(@d.select(Sequel.lit('"name"')).sql).to eq( \
      'SELECT "name" FROM "test"'
    )

    expect(@d.select(Sequel.lit('max(test."name") AS "max_name"')).sql).to eq( \
      'SELECT max(test."name") AS "max_name" FROM "test"'
    )

    expect(@d.insert_sql(:x => :y)).to match( \
      /\AINSERT INTO "test" \("x"\) VALUES \("y"\)( RETURNING NULL)?\z/
    )

  end

  specify "should quote fields correctly when reversing the order if quoting identifiers" do
    expect(@d.reverse_order(:name).sql).to eq( \
      'SELECT * FROM "test" ORDER BY "name" DESC'
    )

    expect(@d.reverse_order(:name.desc).sql).to eq( \
      'SELECT * FROM "test" ORDER BY "name" ASC'
    )

    expect(@d.reverse_order(:name, :test.desc).sql).to eq( \
      'SELECT * FROM "test" ORDER BY "name" DESC, "test" ASC'
    )

    expect(@d.reverse_order(:name.desc, :test).sql).to eq( \
      'SELECT * FROM "test" ORDER BY "name" ASC, "test" DESC'
    )
  end

  specify "should support regexps" do
    @d << {:name => 'abc', :value => 1}
    @d << {:name => 'bcd', :value => 2}

    expect(@d.filter(:name => /bc/).count).to eq(2)
    expect(@d.filter(:name => /^bc/).count).to eq(1)
  end

  specify "#columns should return the correct column names" do
    expect(@d.columns!).to eq([:name, :value])
    expect(@d.select(:name).columns!).to eq([:name])
  end
end


describe "A Vertica dataset with a timestamp field" do
  before do
    @db = VERTICA_DB
    @d = @db[:test3]
    @d.delete if @d.count > 0 # Vertica will throw an error if the table has just been created and does not have a super projection yet.
  end
  after do
    @db.convert_infinite_timestamps = false if @db.adapter_scheme == :postgres
  end

  cspecify "should store milliseconds in time fields for Time objects", :do, :swift do
    t = Time.now
    @d << {:value=>1, :time=>t}
    t2 = @d[:value =>1][:time]
    expect(@d.literal(t2)).to eq(@d.literal(t))
    expect(t2.strftime('%Y-%m-%d %H:%M:%S')).to eq(t.strftime('%Y-%m-%d %H:%M:%S'))
    expect(t2.is_a?(Time) ? t2.usec : t2.strftime('%N').to_i/1000).to eq(t.usec)
  end

  cspecify "should store milliseconds in time fields for DateTime objects", :do, :swift do
    t = DateTime.now
    @d << {:value=>1, :time=>t}
    t2 = @d[:value =>1][:time]
    expect(@d.literal(t2)).to eq(@d.literal(t))
    expect(t2.strftime('%Y-%m-%d %H:%M:%S')).to eq(t.strftime('%Y-%m-%d %H:%M:%S'))
    expect(t2.is_a?(Time) ? t2.usec : t2.strftime('%N').to_i/1000).to eq(t.strftime('%N').to_i/1000)
  end

  describe "Verticas's EXPLAIN and EXPLAIN LOCAL" do
    specify "should not raise errors" do
      @d = VERTICA_DB[:test3]
      expect{@d.explain}.not_to raise_error
      expect{@d.explain(:local => true)}.not_to raise_error
    end
  end

end


describe "A Vertica database" do
  before do
    @db = VERTICA_DB
  end

  specify "should support ALTER TABLE DROP COLUMN" do
    @db.create_table!(:test3) { varchar :name; integer :value }
    expect(@db[:test3].columns).to eq([:name, :value])
    @db.drop_column :test3, :value
    expect(@db[:test3].columns).to eq([:name])
  end

  specify "It does not support ALTER TABLE ALTER COLUMN TYPE" do
    @db.create_table!(:test4) { varchar :name; integer :value }
    expect{ @db.set_column_type :test4, :value, :float }.to raise_error(Sequel::DatabaseError,
                                                    /Syntax error at or near "TYPE"/)
  end

  specify "should support rename column operations" do
    @db.create_table!(:test5) { varchar :name; integer :value }
    @db[:test5] << {:name => 'mmm', :value => 111}
    @db.rename_column :test5, :value, :val
    expect(@db[:test5].columns).to eq([:name, :val])
    expect(@db[:test5].first[:val]).to eq(111)
  end

  specify "should support add column operations" do
    @db.create_table!(:test2) { varchar :name; integer :value }
    expect(@db[:test2].columns).to eq([:name, :value])

    @db.add_column :test2, :xyz, :varchar, :default => '000'
    expect(@db[:test2].columns).to eq([:name, :value, :xyz])
    @db[:test2] << {:name => 'mmm', :value => 111}
    expect(@db[:test2].first[:xyz]).to eq('000')
  end

  specify "#locks should be a dataset returning database locks " do
    expect(@db.locks).to be_a_kind_of(Sequel::Dataset)
    expect(@db.locks.all).to be_a_kind_of(Array)
  end
end

describe "Vertica::Dataset#insert" do
  before do
    @db = VERTICA_DB
    @db.create_table!(:test5){ :xid; Integer :value}
    @db.sqls.clear
    @ds = @db[:test5]
  end

  after do
    @db.drop_table?(:test5)
  end

  specify "should work with static SQL" do
    expect(@ds.with_sql('INSERT INTO test5 (value) VALUES (10)').insert).to eq(1)
    expect(@db['INSERT INTO test5 (value) VALUES (20)'].insert).to eq(1)
    expect(@ds.all).to include({:value=>10}, {:value=>20})
  end

  specify "should insert correctly if using a column array and a value array" do
    expect(@ds.insert([:value], [10])).to eq(1)
    expect(@ds.all).to eq([{:value=>10}])
  end
end

describe "Vertica::Database schema qualified tables" do
  before do
    VERTICA_DB << "CREATE SCHEMA schema_test"
    VERTICA_DB.instance_variable_set(:@primary_keys, {})
    VERTICA_DB.instance_variable_set(:@primary_key_sequences, {})
  end

  after do
    VERTICA_DB << "DROP SCHEMA schema_test CASCADE"
  end

  specify "should be able to create, drop, select and insert into tables in a given schema" do
    VERTICA_DB.create_table(:schema_test__table_in_schema_test){integer :i}
    expect(VERTICA_DB[:schema_test__table_in_schema_test].first).to eq(nil)
    expect(VERTICA_DB[:schema_test__table_in_schema_test].insert(:i=>1)).to eq(1)
    expect(VERTICA_DB[:schema_test__table_in_schema_test].first).to eq({:i=>1})
    expect(VERTICA_DB.from(Sequel.lit('schema_test.table_in_schema_test')).first).to eq({:i=>1})
    VERTICA_DB.drop_table(:schema_test__table_in_schema_test)
    VERTICA_DB.create_table(:table_in_schema_test.qualify(:schema_test)){integer :i}
    expect(VERTICA_DB[:schema_test__table_in_schema_test].first).to eq(nil)
    expect(VERTICA_DB.from(Sequel.lit('schema_test.table_in_schema_test')).first).to eq(nil)
    VERTICA_DB.drop_table(:table_in_schema_test.qualify(:schema_test))
  end

  specify "#tables should not include tables in a default non-public schema" do
    VERTICA_DB.create_table(:schema_test__table_in_schema_test){integer :i}
    expect(VERTICA_DB.tables).to include(:table_in_schema_test)
    expect(VERTICA_DB.tables).not_to include(:tables)
    expect(VERTICA_DB.tables).not_to include(:columns)
    expect(VERTICA_DB.tables).not_to include(:locks)
    expect(VERTICA_DB.tables).not_to include(:domain_udt_usage)
  end

  specify "#tables should return tables in the schema provided by the :schema argument" do
    VERTICA_DB.create_table(:schema_test__table_in_schema_test){integer :i}
    expect(VERTICA_DB.tables(:schema=>:schema_test)).to eq([:table_in_schema_test])
  end

  specify "#schema should not include columns from tables in a default non-public schema" do
    VERTICA_DB.create_table(:schema_test__domains){integer :i}
    sch = VERTICA_DB.schema(:domains)
    cs = sch.map{|x| x.first}
    expect(cs).to include(:i)
    expect(cs).not_to include(:data_type)
  end

  specify "#schema should only include columns from the table in the given :schema argument" do
    VERTICA_DB.create_table!(:domains){integer :d}
    VERTICA_DB.create_table(:schema_test__domains){integer :i}
    sch = VERTICA_DB.schema(:domains, :schema=>:schema_test)
    cs = sch.map{|x| x.first}
    expect(cs).to include(:i)
    expect(cs).not_to include(:d)
    VERTICA_DB.drop_table(:domains)
  end

  specify "#table_exists? should see if the table is in a given schema" do
    VERTICA_DB.create_table(:schema_test__schema_test){integer :i}
    expect(VERTICA_DB.table_exists?(:schema_test__schema_test)).to eq(true)
  end

end
